/**
 *  MOSAIIC Model - 9 decembre 2014
 *  Author: patricktaillandier
 *  Description: 
 */
 
model testTrafficModel
 
global {   
	float critical_concentration_coeff <- 2.0;
	float max_speed <- 50 #km/#h;
	geometry shape <- rectangle(size_road + 20.0,size_road +  20);
	graph road_network_speed;  
	
	float car_size <- 4.0;
	map<int,float> objectives ;
	float g_mean_speed;
	float g_number;
 	list<node_> vertices;
	float size_road <- max_speed * 18 ;
	
	float section_size <- max_speed ;//100#m ; // DeltaX
	bool use_macro_road <- false parameter: true;
	bool save_result <- true;
	
	int nb_car_adds <- 10;
	int nb_cars <- nb_car_adds;
	
	bool record <- false;
	int wait_cpt_max <- 120;
	int wait_cpt;
	int record_cpt_max <- 60;
	int record_cpt;
	
	int interval_cpt_max <- record_cpt_max;
	int interval_cpt; 
	int nb_cars_max <- 160;
	list<people> order_to_active;
	bool new_cycle <-false;
	
	float mean_speed_g;
	bool has_result <- false;
	bool mode_batch <- true;
	
	float P_security_dist_coeff <- 0.9;
	float P_min_security_dist <- 3.0;
	float P_max_accel <- 14.0;
	float error;
	init {
		do initialization;
	}
	
	int cycle_tot_ag<-0;
	int cycle_tot_macro<-0;
	
	float compuation_time;
	
	action initialization {
		if not use_macro_road {
				matrix mat <- matrix(csv_file("objective.csv", ",", float));
			loop i from: 0 to: mat.rows - 1 {
				objectives[int(mat[0,i])] <- float(mat[1,i]);
			}
		}
	
		
		create node_ with:[location:: {10,size_road/2.0+10}, is_traffic_signal::false, is_crossing :: false];
		create node_ with:[location:: {size_road+10,size_road/2.0+10}, is_traffic_signal::false, is_crossing :: false];
		create road with:[shape:: line([first(node_).location, first(node_).location - {0.0,size_road/2.0}, (node_[1]).location-{0.0,size_road/2.0},(node_[1]).location]),highway::"motorway",lanes::1, maxspeed::max_speed] {
			capacite_max <- 1+ int(lanes * shape.perimeter/(4.0));
			
		}
		create road with:[shape:: line([(node_[1]).location,(node_[1]).location + {0.0,size_road/2.0}, first(node_).location+{0.0,size_road/2.0}, first(node_).location]),highway::"motorway",lanes::1, maxspeed::max_speed] {
			capacite_max <- 1+ int(lanes * shape.perimeter/(4.0));
			
		}
		map general_speed_map_speed <- road as_map (each::(each.shape.perimeter / each.maxspeed)); 
		
		
		road_network_speed <-  (as_driving_graph(road, node_))  with_weights general_speed_map_speed;
		
	
		do create_sections;
		
		do add_people(nb_cars);
	}
	
	
	reflex do_not_record when: not record {
		if (new_cycle) {	do initialization; new_cycle <- false;}
		wait_cpt <- wait_cpt + 1;
		if (wait_cpt > wait_cpt_max ) {
			record <- true;
			wait_cpt <- 0;
		}
	}
	
	reflex do_record when: record {
		record_cpt <- record_cpt + 1;
		has_result <- false;
		interval_cpt <- interval_cpt + 1;
		if (not use_macro_road) {
			mean_speed_g <- mean_speed_g + (people mean_of each.real_speed) ;
		}	
		if (interval_cpt = interval_cpt_max) {
			has_result <- true;
			if (not use_macro_road) {
				mean_speed_g <- mean_speed_g / interval_cpt_max;
				error <- error + abs(objectives[nb_cars] - mean_speed_g);
				if save_result {save ""+nb_cars+","+mean_speed_g to:"result_model.csv" rewrite: false;}
			}
			else {
				if save_result {save ""+nb_cars+","+mean_speed_g to:"objective2.csv" rewrite: false;}
			}
			interval_cpt <- 0;
			
		} 
		if (record_cpt > record_cpt_max ) {
			record_cpt <- 0;
			record <- false;
			if (not use_macro_road) {mean_speed_g <- 0.0;}
			nb_cars <- nb_cars + nb_car_adds;
			if (nb_cars > nb_cars_max) {if not mode_batch {do pause;}}
			
			else {
				ask people {do die;}
				ask road {do die;}
				ask section {do die;}
				ask node_ {do die;}
				new_cycle <- true;
			}
		}
	}
	
	reflex oder_def when: length(people) > 0 {
		order_to_active <- [];
		people first_ag <- people with_max_of (each.distance_to_next());
		int nb <- length(people);
		int id <- (first_ag.id);
		loop i from: 0 to: length(people) - 1 {
			order_to_active << people[((i+id) mod nb)];
		}
	}
	
	
	action add_people(int nb_people) {
		if (not use_macro_road) {
			int cpt;
			loop times: nb_people {
				do create_people(cpt);
				cpt <- cpt + 1;
			}
		} else {
			int entier <- int(nb_people/ length(section));
			int reste <- nb_people - entier;
			loop sect over: section {
				loop times: entier{
					ask sect {
						current_concentration <- current_concentration + 1;
					}
				}
			}
			loop times: reste{
				section sect <- one_of(section);
				ask sect {
					current_concentration <- current_concentration + 1;
				}
			}
		}	
		
	}
	
	action create_people(int index)  {
	
		create people  { 
			id <- index;
			min_security_distance <- P_min_security_dist;
			vehicle_length <- car_size ;
			road rd <- first(road);
			location <- first(rd.shape.points);
			target_node <- last(node_);
			security_distance_coeff <- P_security_dist_coeff;  
			max_acceleration <-P_max_accel #km/#h;
			max_speed <- 150 #km /#h ;
			speed_coeff <-1.0;
			ask rd {
				do register(myself,0);
			}
		}
		ask people {
			speed_coeff <- 1.0 + 0.5 * (1 - nb_cars/nb_cars_max);
			max_acceleration <- 100.0;
			
			max_speed <- 300.0;
			do drive_random ;
			speed_coeff <- 1.0;
			max_speed <- 150 #km /#h ;
			max_acceleration <-P_max_accel #km/#h;
			do define_next;
			
		}
		
	}
	
	action create_sections {
		ask road where (each.shape.perimeter >= section_size ) {
			is_macro <- true;
			int nb <- round (shape.perimeter /section_size);
			float sec_length <- shape.perimeter / nb;
			section prev <- nil;
			int cpt <- 0;
			list<point> pts <- points_on (shape,sec_length);
			list<geometry> geoms;
			geometry g <- copy(shape);
			loop pt over: pts {
				if (pt distance_to first(shape.points) > (section_size/10.0)) and (pt distance_to last(shape.points) > (section_size/10.0)) {
					list<geometry> gg <- g split_at pt;
					geoms<< gg[0];
					g <- gg[1];
				}
			}
			geoms << g;
			loop times: nb {
				create section with:[location::location]{
					max_speed <- myself.maxspeed #km/#h;
					color <-  #red;
					previous <- prev;
					if (previous != nil) {previous.next <- self;}
					prev <- self;
					Kmax <- myself.lanes * sec_length /  (car_size) ;
					critical_concentration <- Kmax / critical_concentration_coeff; //Kc
					current_concentration <- 0.0 ;
					current_flow <- current_concentration * max_speed ;
					max_flow <- critical_concentration * max_speed ; 
					myself.sections << self;
					my_road <- myself;
					point pt <- first(node_).location;
					shape <- geoms[cpt];
					cpt <- cpt + 1;
					
				}	
			}
			regular_sections <- sections where ((each.next != nil) and (each.previous != nil));
			ask target_node as: node_{
				inputSections << last(myself.sections);
			}
			ask source_node as: node_{
				outputSections << first(myself.sections);
			}
		} 
	}
	
	
	reflex manage_macro_roads when: use_macro_road{
		float t <- machine_time;
		ask section {
			previous_concentration <- current_concentration;
			do compute_demand_offer;	
		}
		ask node_ {
			do manage_flow;
		}
		
		ask road {
			do manage_flow;
		}
		ask node_ {
			do manage_flow_end;
		}
		
		ask section {
			speed_mean <- previous_concentration = 0 ? 0.0 : current_flow/previous_concentration ;
		}
		compuation_time <- compuation_time + machine_time - t;
		mean_speed_g <- section sum_of(each.speed_mean * each.current_concentration);
		float tot_conc <- section sum_of(each.current_concentration);
		mean_speed_g <- tot_conc> 0 ? (mean_speed_g / tot_conc)  #h/#km : 0.0;
	}
} 



species node_ skills: [skill_road_node] {
	list<section> inputSections;
	list<section> outputSections;
	bool is_crossing;
	bool is_traffic_signal;
	aspect base {    
		draw square(4) color: #magenta;
	} 
	
	action manage_flow {
		float sum_demand <- sum(inputSections collect each.demand);
		float remaining_demand <- copy(sum_demand);
		ask outputSections {
			next_offer <- (self.next).offer ;
			current_flow <- min([self.demand, next_offer]);
			offer_tmp <- 0.0;
		}
		list<section> possible_sections <- outputSections where (each.offer > each.offer_tmp);
		loop while: (remaining_demand > 0) and (not empty(possible_sections)) {
			possible_sections <- outputSections where (each.offer > each.offer_tmp);
			float to_demand <- not empty(possible_sections) ? (remaining_demand / length(possible_sections)) with_precision 2 : 0.0;
			if (to_demand = 0) {break;}
			ask possible_sections {
				float to_give <- min([to_demand, offer - offer_tmp]);
				offer_tmp <- offer_tmp + to_give;
				remaining_demand <- remaining_demand - to_give;
			}	
		}
		ask outputSections {
			current_concentration <- current_concentration + step/section_size *(offer_tmp - self.current_flow) ;
		}
		
		ask inputSections {
			current_flow <- 0.0;
		}
		possible_sections <- inputSections where (each.demand > each.current_flow);
		float real_demand <- sum_demand - remaining_demand;
		loop while: (real_demand > 0) and (not empty(possible_sections)) {
			possible_sections <- inputSections where (each.demand > each.current_flow);
			float to_demand <- not empty(possible_sections) ? (real_demand / length(possible_sections))with_precision 2 : 0.0;
			if (to_demand = 0) {break;}
			ask possible_sections {
				float to_give <- min([to_demand, demand - current_flow]);
				current_flow <- current_flow + to_give;
				real_demand <- real_demand - to_give;
			}	
		}
		
				
	}
	
	action manage_flow_end{
		ask inputSections {
			float previous_flow <- (self.previous).current_flow;
			current_concentration <- current_concentration + step/section_size *(previous_flow - self.current_flow) ;
		}
	}
}

species road skills: [skill_road] {
	string oneway;
	int nb_people <- 0;
	int nb_people_tot <- 0;
	
	int capacite_max ;
	float nb_bloques <- 0.0;
	int min_traffic_jam_people_destroy;
	int min_traffic_jam_people_creation;
	
	road next_linked_road;
	string highway;
	list<node_> neighbours;
	
	list<section> sections;
	bool is_macro <- false;
			
	list<people> cars_to_capture <- [];
	list<road> next_roads;
	list<section> regular_sections;
	
	aspect carto {
		if not(use_macro_road) {
			draw shape color: #black;
		} else {
			loop sect over: sections {
				draw sect.shape + (1 + 3 * sect.current_concentration/sect.critical_concentration ) border: #black color: #red;
			}
		}
	}
	
	
	action manage_flow  {
		ask regular_sections{ 
			next_offer <- (self.next).offer ;
			current_flow <- min([self.demand, next_offer]);
		}
		ask regular_sections {
			float previous_flow <- (self.previous).current_flow;
			current_concentration <- current_concentration + step/section_size *(previous_flow - self.current_flow) ;
			
		}
	}
	
}

species section schedules:[]{
	float current_concentration  ;
	float current_flow  ;
	float previous_concentration;
	road my_road;
	float offer ;
	float demand  ;
	bool is_last;
	rgb color;
	float remaining_car;
	float Kmax;
	float critical_concentration;
	float max_speed ; // Vl
	float max_flow ; //Qx
	float next_offer;
	float offer_tmp;

	float speed_mean;
	section previous ;
	section next ;
	
	geometry shape;
	
	action compute_demand_offer{
		next_offer <- 0.0;
		if current_concentration <= critical_concentration {
			offer <- max_flow ;
			demand <- max([0 , max_flow/critical_concentration * current_concentration ]);
		} else {
			 offer <- max([0 , - max_flow/critical_concentration * current_concentration + 2 * max_flow]);
			 demand <- max_flow ;
		}
	}
}

	
species people skills: [advanced_driving] schedules: order_to_active {
	rgb color <-rnd_color(255) ;
	node_ target_node;
	int id;
	bool is_stopped <- false;
	bool recompute_path <- false;
	bool to_delete <- false;
	float proba_avoid_traffic_jam;
	node_ current_node;
	bool mode_avoid <- false;
	int cpt_avoid <- 0;
	people next;
	action remove_from_simu {
		if (current_edge != nil) {
			ask road(current_edge) {
				do unregister agent: myself;
			}
		}
		do die;
	}
	reflex driving {
		float t <- machine_time;
		do drive_random;
		compuation_time <- compuation_time + machine_time - t;
	} 
	
	action define_next {
		int index <- (id + 1);
		if (index = length(people)) {index <- 0;}
		next <- people[id];
		
	}
	
	float distance_to_next{
		return next distance_to self;
	}
	
	aspect base { 
		draw triangle(4) color: color rotate:heading + 90;	
	} 
	

} 
experiment traffic_simulation type: gui {
	
	action _init_ {
		create testTrafficModel_model with: [mode_batch::false];
	}
	output {
	monitor error value: error;
	display city_display2D {
			species road aspect: carto ;
			species node_ aspect: base;
			species people aspect: base; 
			
		}
		
		display phase_diagram refresh: has_result    {
			chart "concentration_speed" type: xy /* size:{0.5,0.5} position:{0.0,0.0} */ x_label: 'Concentration (nb cars)' y_label: 'Vitesse (km/h)'
			{
				data legend: "concentration_speed" value: [nb_cars, mean_speed_g #h/#km,1]  color: #green style: dot;
			}
	
		}
	}
	
}


